(****** Locales/--AMOK-Header-- *********************************************

:Program.    Locales.mod
:Contents.   abstract class Locale for string localisation handling classes
:Author.     Oliver Knorr
:Copyright.  Public Domain
:Language.   Oberon-2
:Translator. Amiga Oberon v3.11
:History.    v1.0 [olk] 21-Mar-94
:Remark.     needs Interfaces 40.15 or better to compile
:Version.    $VER: Locales.mod 1.0 (21.3.94)

*****************************************************************************
*
*)

MODULE Locales;

(****** Locales/--background-- **********************************************
*
*       This module defines the abstract class Locale with one method
*       GetString() thought for retrieving localized strings for an
*       application. Oberon-2 modules generated automatically by CatCompMO
*       always define a subclass of Locale that redefines GetString()
*       to a working method for the translated catalog description and
*       export an object of this class as a global variable.
*
*       The idea behind the Locale class is to give module packages
*       not written for one special project access to an applications
*       localized strings. To achieve this, the module package has to
*       import Locales.Locale and to access the strings via the GetString()
*       method of a Locale object that can be given to the module package
*       with an initialization procedure. This way it does not need to
*       know the name of the module that defines the locale strings. Your
*       main module simply passes the global object that is exported in
*       the module generated by CatCompMO to the module package.
*
*****************************************************************************
*
*)

IMPORT

  BasicTypes,
  Exec;

TYPE

  Locale * = POINTER TO LocaleDesc;
  LocaleDesc * = RECORD (BasicTypes.ANYDesc) END;


(****** Locales/Locale.GetString ********************************************
*
*   NAME
*       Locale.GetString -- get localized string  (method)
*
*   SYNOPSIS
*       (l: Locale) GetString (ID: LONGINT): Exec.LSTRPTR
*
*   FUNCTION
*       *** absctract method ***
*       Must be redefined by subclasses to return the localized
*       string with the given ID.
*
*   INPUTS
*       ID - locale ID for the string that shall be returned
*
*   RESULT
*       An Exec.LSTRPTR to the localized string.
*
*   EXAMPLE
*
*   NOTES
*
*   BUGS
*
*   SEE ALSO
*
*****************************************************************************
*
*)

PROCEDURE (l: Locale) GetString * (ID: LONGINT): Exec.LSTRPTR;
END GetString;

END Locales.
